@@ -1,4 +1,4 @@
-# Copyright 2004-2015 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -43,7 +43,7 @@ from renpy.display.predict import displayable as predict_displayable
 
 from renpy.python import py_eval_bytecode
 from renpy.pyanalysis import Analysis, NOT_CONST, GLOBAL_CONST, ccache
-import md5
+import hashlib
 
 # This file contains the abstract syntax tree for a screen language
 # screen.
@@ -61,6 +61,7 @@ filename = '<screen language>'
 # A log that's used for profiling information.
 profile_log = renpy.log.open("profile_screen", developer=True, append=False, flush=False)
 
+
 def compile_expr(node):
     """
     Wraps the node in a python AST, and compiles it.
@@ -96,7 +97,7 @@ class SLContext(renpy.ui.Addable):
         self.keywords = { }
 
         # The style prefix that is given to children of this displayable.
-        self.style_prefix = ""
+        self.style_prefix = None
 
         # A cache associated with this context. The cache maps from
         # statement serial to information associated with the statement.
@@ -147,17 +148,6 @@ class SLContext(renpy.ui.Addable):
         # to speed things up.
         self.unlikely = False
 
-
-    def get_style_group(self):
-        style_prefix = self.style_prefix
-
-        if style_prefix:
-            return style_prefix[:-1]
-        else:
-            return None
-
-    style_group = property(get_style_group)
-
     def add(self, d, key):
         self.children.append(d)
 
@@ -214,7 +204,6 @@ class SLNode(object):
 
         raise Exception("copy not implemented by " + type(self).__name__)
 
-
     def report_traceback(self, name, last):
         if last:
             return None
@@ -299,6 +288,7 @@ class SLNode(object):
 # A sentinel used to indicate a keyword argument was not given.
 NotGiven = renpy.object.Sentinel("NotGiven")
 
+
 class SLBlock(SLNode):
     """
     Represents a screen language block that can contain keyword arguments
@@ -351,7 +341,7 @@ class SLBlock(SLNode):
                 keyword_values[k] = py_eval_bytecode(compile_expr(node))
             else:
                 keyword_keys.append(ast.Str(s=k))
-                keyword_exprs.append(node) # Will be compiled as part of ast.Dict below.
+                keyword_exprs.append(node)  # Will be compiled as part of ast.Dict below.
 
             self.constant = min(self.constant, const)
 
@@ -379,7 +369,6 @@ class SLBlock(SLNode):
                 self.last_keyword = True
                 break
 
-
     def execute(self, context):
 
         # Note: SLBlock.execute() is inlined in various locations for performance
@@ -408,12 +397,13 @@ class SLBlock(SLNode):
         for i in self.keyword_children:
             i.keywords(context)
 
-        style_group = context.keywords.pop("style_group", NotGiven)
-        if style_group is not NotGiven:
-            if style_group is not None:
-                context.style_prefix = style_group + "_"
-            else:
-                context.style_prefix = ""
+        style_prefix = context.keywords.pop("style_prefix", NotGiven)
+
+        if style_prefix is NotGiven:
+            style_prefix = context.keywords.pop("style_group", NotGiven)
+
+        if style_prefix is not NotGiven:
+            context.style_prefix = style_prefix
 
     def copy_on_change(self, cache):
         for i in self.children:
@@ -433,6 +423,7 @@ class SLBlock(SLNode):
 
 list_or_tuple = (list, tuple)
 
+
 class SLCache(object):
     """
     The type of cache associated with an SLDisplayable.
@@ -489,6 +480,7 @@ class SLCache(object):
 # the parent.
 NO_DISPLAYABLE = renpy.display.layout.Null()
 
+
 class SLDisplayable(SLBlock):
     """
     A screen language AST node that corresponds to a displayable being
@@ -497,6 +489,9 @@ class SLDisplayable(SLBlock):
 
     hotspot = False
 
+    # A list of variables that are locally constant.
+    local_constant = [ ]
+
     def __init__(self, loc, displayable, scope=False, child_or_fixed=False, style=None, text_style=None, pass_context=False, imagemap=False, replaces=False, default_keywords={}, hotspot=False):
         """
         `displayable`
@@ -584,6 +579,11 @@ class SLDisplayable(SLBlock):
         if self.imagemap:
             analysis.pop_control()
 
+        # If we use a scope, store the local constants that need to be
+        # kept and placed into the scope.
+        if self.scope:
+            self.local_constant = list(analysis.local_constant)
+
     def prepare(self, analysis):
 
         SLBlock.prepare(self, analysis)
@@ -606,7 +606,7 @@ class SLDisplayable(SLBlock):
                 has_values = True
             else:
                 values.append(use_expression)
-                exprs.append(node) # Will be compiled as part of the tuple.
+                exprs.append(node)  # Will be compiled as part of the tuple.
                 has_exprs = True
 
             self.constant = min(self.constant, const)
@@ -658,13 +658,20 @@ class SLDisplayable(SLBlock):
 
         if cache.constant and (cache.style_prefix == context.style_prefix):
 
-            for i in cache.constant_uses_scope:
+            for i, local_scope in cache.constant_uses_scope:
+
+                if local_scope:
+                    scope = dict(context.scope)
+                    scope.update(local_scope)
+                else:
+                    scope = context.scope
+
                 if copy_on_change:
-                    if i._scope(context.scope, False):
+                    if i._scope(scope, False):
                         cache.constant = None
                         break
                 else:
-                    i._scope(context.scope, True)
+                    i._scope(scope, True)
 
             else:
 
@@ -725,6 +732,7 @@ class SLDisplayable(SLBlock):
 
             arguments = keywords.pop("arguments", None)
             properties = keywords.pop("properties", None)
+            style_suffix = keywords.pop("style_suffix", None) or self.style
 
             if arguments:
                 positional += arguments
@@ -733,8 +741,11 @@ class SLDisplayable(SLBlock):
                 keywords.update(properties)
 
             # If we don't know the style, figure it out.
-            if ("style" not in keywords) and self.style:
-                keywords["style"] = ctx.style_prefix + self.style
+            if ("style" not in keywords) and style_suffix:
+                if ctx.style_prefix is None:
+                    keywords["style"] = style_suffix
+                else:
+                    keywords["style"] = ctx.style_prefix + "_" + style_suffix
 
             if widget_id and (widget_id in screen.widget_properties):
                 keywords.update(screen.widget_properties[widget_id])
@@ -770,7 +781,7 @@ class SLDisplayable(SLBlock):
                 # the one that gets the scope, and gets children added to it.)
                 main = old_main
 
-                if widget_id:
+                if widget_id and not ctx.unlikely:
                     screen.widgets[widget_id] = main
 
                 if self.scope and main._uses_scope:
@@ -806,14 +817,14 @@ class SLDisplayable(SLBlock):
 
                 main._location = self.location
 
-                if widget_id:
+                if widget_id and not ctx.unlikely:
                     screen.widgets[widget_id] = main
                 # End child creation code.
 
                 imagemap = self.imagemap
 
-                cache.copy_on_change = False # We no longer need to copy on change.
-                cache.children = None # Re-add the children.
+                cache.copy_on_change = False  # We no longer need to copy on change.
+                cache.children = None  # Re-add the children.
 
             if debug:
                 if reused:
@@ -921,20 +932,31 @@ class SLDisplayable(SLBlock):
 
         if (transform is not None) and (d is not NO_DISPLAYABLE):
             if reused and (transform == cache.raw_transform):
+
+                if isinstance(cache.transform, renpy.display.transform.Transform):
+                    if cache.transform.child is not d:
+                        cache.transform.set_child(d, duplicate=False)
+
                 d = cache.transform
             else:
                 cache.raw_transform = transform
 
                 if isinstance(transform, Transform):
                     d = transform(child=d)
+                    d._unique()
+
                 elif isinstance(transform, list_or_tuple):
                     for t in transform:
                         if isinstance(t, Transform):
                             d = t(child=d)
                         else:
                             d = t(d)
+
+                        d._unique()
+
                 else:
                     d = transform(d)
+                    d._unique()
 
                 if isinstance(d, Transform):
                     old_transform = cache.transform
@@ -959,7 +981,14 @@ class SLDisplayable(SLBlock):
                 cache.constant = d
 
                 if self.scope and main._uses_scope:
-                    ctx.uses_scope.append(main)
+
+                    local_scope = { }
+
+                    for i in self.local_constant:
+                        if i in ctx.scope:
+                            local_scope[i] = ctx.scope[i]
+
+                    ctx.uses_scope.append((main, local_scope))
 
                 cache.constant_uses_scope = ctx.uses_scope
 
@@ -1201,6 +1230,7 @@ class SLIf(SLNode):
 
         return False
 
+
 class SLShowIf(SLNode):
     """
     The AST node that corresponds to the showif statement.
@@ -1250,7 +1280,10 @@ class SLShowIf(SLNode):
 
     def execute(self, context):
 
-        first_true = True
+        # This is true when the block should be executed - when no outer
+        # showif is False, and when no prior block in this showif has
+        # executed.
+        first_true = context.showif is not False
 
         for cond, block in self.prepared_entries:
 
@@ -1320,7 +1353,6 @@ class SLFor(SLBlock):
 
         analysis.pop_control()
 
-
     def prepare(self, analysis):
         node = ccache.ast_eval(self.expression)
 
@@ -1341,7 +1373,6 @@ class SLFor(SLBlock):
 
     def execute(self, context):
 
-
         variable = self.variable
         expr = self.expression_expr
 
@@ -1439,6 +1470,7 @@ class SLPass(SLNode):
 
         return rv
 
+
 class SLDefault(SLNode):
 
     def __init__(self, loc, variable, expression):
@@ -1623,7 +1655,6 @@ class SLUse(SLNode):
             if cache is None:
                 context.cache[self.serial] = cache = { }
 
-
         # Evaluate the arguments.
         try:
             if self.args:
@@ -1693,7 +1724,6 @@ class SLTransclude(SLNode):
     def __init__(self, loc):
         SLNode.__init__(self, loc)
 
-
     def copy(self, transclude):
         rv = self.instantiate(transclude)
         rv.constant = transclude
@@ -1719,7 +1749,12 @@ class SLTransclude(SLNode):
         ctx.showif = context.showif
         ctx.uses_scope = context.uses_scope
 
-        context.transclude.execute(ctx)
+        try:
+            renpy.ui.stack.append(ctx)
+            context.transclude.keywords(ctx)
+            context.transclude.execute(ctx)
+        finally:
+            renpy.ui.stack.pop()
 
         if ctx.fail:
             context.fail = True
@@ -1744,7 +1779,6 @@ class SLScreen(SLBlock):
 
     version = 0
 
-
     # This screen's AST when the transcluded block is entirely
     # constant (or there is no transcluded block at all). This may be
     # the actual AST, or a copy.
@@ -1757,6 +1791,7 @@ class SLScreen(SLBlock):
     # The analysis
     analysis = None
 
+    layer = "'screens'"
 
     def __init__(self, loc):
 
@@ -1775,10 +1810,10 @@ class SLScreen(SLBlock):
         self.tag = None
 
         # The variant of screen we're defining.
-        self.variant = "None" # expr.
+        self.variant = "None"  # expr.
 
         # Should we predict this screen?
-        self.predict = "None" # expr.
+        self.predict = "None"  # expr.
 
         # The parameters this screen takes.
         self.parameters = None
@@ -1822,6 +1857,7 @@ class SLScreen(SLBlock):
             predict=renpy.python.py_eval(self.predict),
             parameters=self.parameters,
             location=self.location,
+            layer=renpy.python.py_eval(self.layer),
             )
 
     def analyze(self, analysis):
@@ -1850,7 +1886,6 @@ class SLScreen(SLBlock):
             self.not_const_ast = self.const_ast
             targets = [ self.const_ast ]
 
-
         for ast in targets:
             analysis = ast.analysis = Analysis(None)
 
@@ -1932,6 +1967,9 @@ class SLScreen(SLBlock):
 
         current_screen = renpy.display.screen.current_screen()
 
+        if current_screen.screen_name[0] in renpy.config.profile_screens:
+            debug = True
+
         context = SLContext()
         context.scope = scope
         context.globals = renpy.python.store_dicts["store"]
@@ -1940,6 +1978,7 @@ class SLScreen(SLBlock):
         context.updating = (current_screen.phase == renpy.display.screen.UPDATE)
 
         name = scope["_name"]
+
         main_cache = current_screen.cache
 
         cache = main_cache.get(name, None)
@@ -1954,6 +1993,7 @@ class SLScreen(SLBlock):
         for i in context.children:
             renpy.ui.implicit_add(i)
 
+
 class ScreenCache(object):
 
     def __init__(self):
@@ -1968,14 +2008,15 @@ scache = ScreenCache()
 
 CACHE_FILENAME = "cache/screens.rpyb"
 
+
 def load_cache():
-    if renpy.game.args.compile: # @UndefinedVariable
+    if renpy.game.args.compile:  # @UndefinedVariable
         return
 
     try:
         f = renpy.loader.load(CACHE_FILENAME)
 
-        digest = f.read(md5.digest_size)
+        digest = f.read(hashlib.md5().digest_size)
         if digest != renpy.game.script.digest.digest():
             return
 
@@ -1990,10 +2031,14 @@ def load_cache():
     except:
         pass
 
+
 def save_cache():
     if not scache.updated:
         return
 
+    if renpy.macapp:
+        return
+
     try:
         data = zlib.compress(dumps(scache, 2), 9)
 
@@ -2002,5 +2047,3 @@ def save_cache():
             f.write(data)
     except:
         pass
-
-
